/***********************************************************************\
* NLED Text Editor Copyright 1987-2003, Evan Alexander Weaver           *
*                                                                       *
* This source code, and the associated executable programs obtained by  *
* compiling it, are published under the GNU General Public License. See *
* the accompanying file "nled.txt" for the details of this license and  *
* some other comments about NLED and this source code.                  *
*                                                                       *
* File: NLED.C                            Date Last Modified: 15-Feb-02 *
*                                                                       *
* Main modules for NLED (Neat Little EDitor).                           *
* Functions are:                                                        *
*   main()                                                              *
*   editline() - processes keys that edit the screen                    *
*   editor() - controls all edit mode actions                           *
*   escapes() - processes escape sequences                              *
*   finalize() - undoes initialize()                                    *
*   help() - displays online help screen                                *
*   initialize() - performs required initializations                    *
*                                                                       *
* 05-Oct-91: ANSI function headers instead of "old-style"               *
* 12-Feb-92: UNIX standardization                                       *
* 05-Apr-94: Tab compression, key alternatives, status line control     *
* 02-Sep-94: Default Tab compress for UNIX, nocompress for DOS          *
* 09-Jul-96: Win95 port                                                 *
* 04-Feb-02: 2.5 cleanup release                                        *
* 15-Feb-02: Use MAX_FILE for file name lengths                         *
\***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "nled.h"
#include "nledkeys.h"
#include "nledfdef.h"
#ifdef SHAREWARE
extern void shareware(void);
#endif

/***********************************************************************\
* External variable definitions.                                        *
*                                                                       *
* 14-Dec-88: Init func_mac array to useful defaults                     *
* 13-Feb-92: Added case sensitivity                                     *
* 02-Feb-02: Made filename string much longer                           *
\***********************************************************************/
#if ZORTECH
    int _okbigbuf = 0;  /* tell Zortech to save room for child process */
    #define OPTION '/'
#elif WIN95 || __OS2__
    #define OPTION '/'
#else
    #define OPTION '-'
#endif

int text_mode = TEXT;   /* file mode: text (TEXT or TEXTZ) or binary (FALSE)*/
int ins_mode = TRUE;    /* edit mode: insert (TRUE) or overstrike (FALSE)   */
int beep_level = 2;     /* severity at which to beep (0 - 2)                */
int tab_size = 8;       /* number of spaces per tab stop (1 - 15)           */
#if UNIX
int tab_compress = TRUE;  /* compress spaces to tabs on saving (TRUE/FALSE) */
#else
int tab_compress = FALSE; /* compress spaces to tabs on saving (TRUE/FALSE) */
#endif
int case_sens = FALSE;  /* searches with Find case sensitive? (TRUE/FALSE)  */
int color_setting = -1; /* color setting, -1: use what is in nledcfg.bin    */
                        /*                 0: let display routines select   */
int edit_flag = FALSE;  /* file has (TRUE) or has not (FALSE) been edited   */
                        /* since it was last saved                          */
int status_flag = TRUE;
#if UNIX
int status_disable = FALSE;     /* status line complexity (UNIX -s option)  */
#endif

/* function key macros: plain fkeys in 0-9, Shift   */
/* fkeys in 10-19, Ctl fkeys in 20-29, Alt in 30-39 */
/* [0] is the number of keystrokes, and [1] - [9]   */
/* are the actual keystrokes (up to 9)              */
#if UNIX
int func_mac[40][10] =
#else
unsigned short func_mac[40][10] =
#endif
    {   
        {2, ESC, '?'},
        {0},
        {2, ESC, 'g'},
        {5, CR, CLFT, HOME, CRGT, DOWN},
        {2, ESC, 'b'},
        {2, ESC, 'e'},
        {3, ESC, 'f', CR},
        {9, ESC, 'b', RGHT, ESC, 'e', RGHT, ESC, 'm', 0XFF6D}, /* Alt-F6 */
        {2, ESC, 'l'},
        {2, ESC, 'q'},
        {0},
        {0},
        {5, LEFT, LEFT, LEFT, LEFT, LEFT},
        {5, RGHT, RGHT, RGHT, RGHT, RGHT},
        {2, ESC, 'c'},
        {9, HOME, ESC, 'b', END, ESC, 'e', CR, ESC, 'c'},
        {2, ESC, 'r'},
        {3, ESC, 'o', CR},
        {2, ESC, 'n'},
        {2, ESC, 's'},
        {7, ESC, 'e', CLFT, ESC, 'b', ESC, 'd'},
        {7, ESC, 'b', CRGT, ESC, 'e', ESC, 'd'},
        {5, BS, BS, BS, BS, BS},
        {5, DEL, DEL, DEL, DEL, DEL},
        {2, ESC, 'd'},
        {9, HOME, ESC, 'b', END, ESC, 'e', ESC, 'd', DEL},
        {7, ESC, 'e', HOME, ESC, 'b', ESC, 'd'},
        {7, ESC, 'b', END, ESC, 'e', ESC, 'd'},
        {3, ESC, 'h', 't'},
        {3, ESC, 'h', 'b'},
        {3, ESC, 'p', 0xFF3C}, /* F2 */
        {3, ESC, 'z', RGHT},
        {6, CHOM, UP, UP, UP, UP, UP},
        {6, CEND, DOWN, DOWN, DOWN, DOWN, DOWN},
        {2, ESC, 'm'},
        {2, ESC, 'u'},
        {4, ESC, 'f', 'x', ESC},
        {2, ESC, 'o'},
        {2, ESC, 'i'},
        {2, ESC, 'x'}
    };
int func_key = -1;      /* function key currently executing (0 - 39) or -1  */
int func_ind;           /* keystroke of func_key to be processed next       */

char file_name[MAX_FILE+1];    /* name of current file                      */

INDEX len_ds = 0;   /* # of chars in data string (>= 0)             */
INDEX nol_ds = 1;   /* # of lines in data string (>=1)              */
int row_no = 0;     /* current row on screen (0 - SCRHI-2)          */
int col_no = 0;     /* current col on screen (0 - SCRWI-1)          */
INDEX scr_top = 0;  /* line # of 1st row on screen (0 - nol_ds-1)   */
INDEX scr_left = 0; /* actual col # of screen column 0 (>=0)        */
INDEX abs_pos = 0;  /* current position in data (0 - len_ds)        */
INDEX scr_loff[MAX_ROWS - 1] = {0};
                    /* offset into data of start of each line       */
                    /* currently on the screen (0 - len_ds)         */
INDEX scr_llen[MAX_ROWS - 1] = {0};
                    /* length of each line currently on the screen  */
                    /* not including trailing LF (0 - len_ds)       */


int draw_brow;      /* screen row to begin redrawing        */
int draw_erow = -1;      /* screen row to end redrawing  */
int draw_col;       /* col of draw_brow to begin redrawing  */
int num_rows;       /* number of actual screen rows         */
int num_cols;       /* number of actual screen columns      */

INDEX block_beg;    /* offset in data of start of user marked block */
INDEX block_end;    /* 1+offset in data of end of user marked block */

char prompt_msg[MAX_PM+1];  /* status line message to be displayed  */
int msg_flag;   /* prompt_msg should (TRUE) or should not (FALSE)   */
                /* be displayed                             */
int msg_col;    /* column to position cursor on status line after   */
                /* displaying prompt_msg (0 - num_cols-1) or -1     */

/* globals only needed in this source file: */
int scr_type = -1;     /* type not specified, let display routine select */
char config_file[MAX_FILE+1];            /* default configuration file   */

/***********************************************************************\
* Main function: returns 1 if there was an error                        *
* 01-May-87: added -sX option.                                          *
* 05-Oct-91: return 0 if normal completion                              *
* 17-Feb-92: color, new default configuration file                      *
* 02-Feb-02: longer file name                                           *
* 04-Feb-02: use search path for config file in Windows                 *
\***********************************************************************/
int main(int argc, char *argv[])
{
    int i;

    /* no file loaded yet */
    strcpy(file_name,"");

    /* find default configuration file name */
#if UNIX
    /* look in user's HOME directory */
    strcpy(config_file, getenv("HOME"));
    strcat(config_file, "/nledcfg.bin");
#elif WIN95
    /* look in search path */
    _searchenv("nledcfg.bin", "PATH", config_file);
#else
    /* look in same directory that NLED was found in */
    strcpy(config_file, argv[0]);
    for (i = strlen(config_file); (i > 0) && (config_file[i] != '\\'); i--)
        /* find trailing backslash */;
    config_file[i] = '\0';
    strcat(config_file, "\\nledcfg.bin");
#endif

    i = 0;
    while (++i < argc) {
        if (OPTION == argv[i][0]) {
            switch(argv[i][1]) {
            case 's':
            case 'S':
#if UNIX
                /* make status line blank to speed things up */
                status_disable = TRUE;
#else
                switch (argv[i][2]) {
                case '0':
                    scr_type = 0;
                    break;

                case '1':
                    scr_type = 1;
                    break;

                case '2':
                    scr_type = 2;
                    break;
                }
#endif
                break;

            case 'c':
            case 'C':
                /* note: htoc returns -1 if the number is no good! */
                /* this is good, since -1 means: use config file colors */
                color_setting = htoc(argv[i] + 2);
                break;
            }
        }
        else {
            strncpy(file_name,argv[i],MAX_FILE);
            file_name[MAX_FILE] = '\0';
            break;
        }
    }
    if (initialize()) {
        finalize();
        exit(1);
    }
    editor();
    if (finalize()) exit(1);
    return 0;
}

/***********************************************************************\
* Processes keys which actually edit the file. Doesn't return anything  *
* meaningful.                                                           *
\***********************************************************************/
void editline(int c /* keycode to be processed */)
{
    int n;

    switch (c) {
    case BS:
        if (scr_top + scr_left + col_no + row_no > 0) {
            if (col_no + scr_left > 0) {
                dispupd('c',-1);
                if (scr_left + col_no < scr_llen[row_no]) dispdelc();
            }
            else {
                dispupd('r',-1);
                dispupd('c',(scr_llen[row_no] - scr_left - col_no));
                dispdelc();
            }
        }
        break;

    case HEX:
        if (0 <= (c = inputhex())) {
            if (ins_mode) dispinsc(c,1);
            else            disprepc(c);
            if (c == LF) {
                dispupd('c',-scr_left - col_no);
                dispupd('r',1);
            }
            else dispupd('c',1);
        }
        break;

    case ' ':
        if (scr_llen[row_no] > scr_left + col_no) {
            if (ins_mode) dispinsc(' ',1);
            else          disprepc(' ');
        }
        dispupd('c',1);
        break;

    case CR:
    case LF:
        if (ins_mode)
            dispinsc(LF,1);
        else if (row_no + scr_top + 1 >= nol_ds) {
            dispupd('c',scr_llen[row_no] - scr_left - col_no);
            dispinsc(LF,1);
        }
        dispupd('c',-scr_left - col_no);
        dispupd('r',1);
        break;

    case DEL:
        dispdelc();
        break;

    case TAB:
        n = tab_size*(1+(int)((scr_left+col_no)/tab_size))-scr_left-col_no;
        if (ins_mode && (scr_llen[row_no] > scr_left + col_no)) {
            dispinsc(' ', n);
        }
        dispupd('c', n);
        break;

    default:
        if ((c > ' ') && (c <= LASTPCHAR)) {
            if (ins_mode) dispinsc(c,1);
            else          disprepc(c);
            dispupd('c',1);
        }
        else beep(2);
        break;
    }
}

/***********************************************************************\
* Controls editing mode by processing keys as they are pressed. This    *
* function processes cursor movement key directly, and calls escapes()  *
* and editline() to process other keys. Returns OK when done (or ERR if *
* there was a fatal error which we don't currently have)                *
* 05-Apr-94: added ALT_ key alternatives (for UNIX)                     *
* 20-Mar-95: fixed typo in ALT_STAB #ifdef                              *
\***********************************************************************/
void editor(void)
{
    int c, fk, fknest = 0;

    while (TRUE) {
        c = inputc();

        switch (c) {
        case DOWN:
            dispupd('r',1);
            break;

        case UP:
            dispupd('r',-1);
            break;

        case RGHT:
            dispupd('c',1);
            break;

        case LEFT:
            dispupd('c',-1);
            break;

        case HOME:
#ifdef ALT_HOME
        case ALT_HOME:
#endif
            dispupd('c',-col_no - scr_left);
            break;

        case CHOM:
            dispupd('r',-row_no);
            break;

        case CPGD:
            dispupd('r',(nol_ds - scr_top - 1 - row_no));
            dispupd('c',(scr_llen[row_no] - scr_left - col_no));
            break;

        case END:
#ifdef ALT_END
        case ALT_END:
#endif
            dispupd('c',(scr_llen[row_no] - scr_left - col_no));
            break;

        case CEND:
            dispupd('r',num_rows - 2 - row_no);
            break;

        case PGUP:
#ifdef ALT_PGUP
        case ALT_PGUP:
#endif
            dispupd('t',2-num_rows);
            break;

        case PGDN:
#ifdef ALT_PGDN
        case ALT_PGDN:
#endif
            dispupd('t',num_rows-2);
            break;

        case CLFT:
            dispprevw();
            break;

        case CRGT:
            dispnextw();
            break;

        case CPGU:
            dispupd('t',-scr_top);
            dispupd('l',-scr_left);
            dispupd('r',-row_no);
            dispupd('c',-col_no);
            break;

        case INS:
#ifdef ALT_INS
        case ALT_INS:
        case ALT_INS1:
#endif
            ins_mode = !ins_mode;
            break;

        case STAB:
#ifdef ALT_STAB
        case ALT_STAB:
#endif
            dispupd('c',
             tab_size*(int)((scr_left+col_no-1)/tab_size)-scr_left-col_no);
            break;

        case ESC:
            if (escapes()) return;
            break;

        default:
            if (0 <= (fk = funckey(c))) {
                if (func_key < 0)
                    fknest = 0;
                else
                    fknest++;
                if (fknest < 20) {
                    func_key = fk;
                    func_ind = 0;
                }
                else {
                    /* Note function keys can chain at most 20 deep */
                    func_key = -1;
                    dispmess(" Function key aborted: nested too deep",-1,TRUE);
                    beep(2);
                }
            }
            else
                editline(c);
            break;
        }

        display();
    }
}

/***********************************************************************\
* Handles processing of escape sequences. Returns ERR if program        *
* termination is requested, and OK otherwise                            *
* 01-APR-87: Removed unused variable "d"                                *
* 22-APR-88: Added option Z                                             *
* 13-Feb-92: Added option Y                                             *
* 15-Feb-02: pass NULL to namefile instead of "" (edit,not start over)  *
\***********************************************************************/
int escapes(void)
{
    int again = 1, retcode = OK, c;

    while (again) {
        dispmess(" Command?   (? for help, <Esc> to cancel) ",10,FALSE);

        c = inputc();
        c = toupper(c);
        again = 0;
        msg_flag = FALSE;

        switch (c) {
        case 'A':
            setbeep();
            break;

        case 'B':
            if ((block_beg < block_end) || (abs_pos < block_end)) {
                if ((block_beg >= scr_loff[row_no]) && 
                 (block_beg <= scr_loff[row_no] + scr_llen[row_no])) {
                    setdraw();
                    draw_col = 0;
                }
                else
                    setdrawall();
            }
            block_beg = abs_pos;
            dispmess(" Beginning of block marked",-1,FALSE);
            break;

        case 'C':
            dispcpyblk();
            break;

        case 'D':
            dispdelblk();
            break;

        case 'E':
            if ((block_beg < block_end) || (block_beg < abs_pos)) {
                if ((block_end >= scr_loff[row_no]) && 
                 (block_end <= scr_loff[row_no] + scr_llen[row_no])) {
                    setdraw();
                    draw_col = 0;
                }
                else
                    setdrawall();
            }
            block_end = abs_pos;
            dispmess(" End of block marked",-1,FALSE);
            break;

        case 'F':
            findrepl(0);
            break;

        case 'G':
            gotoline();
            break;

        case 'H':
            setmode();
            break;

        case 'I':
            insfile();
            break;

        case 'K':
            tab_compress = !tab_compress;            
            if (tab_compress)
                dispmess(" Text mode tab compression activated", -1, FALSE);
            else
                dispmess(" Text mode tab compression deactivated", -1, FALSE);
            break;

        case 'L':
            if ((ESCHIT == saveyn()) ||
             (ESCHIT == namefile(" Load file:", NULL)))
                dispmess(" No file loaded ",-1,FALSE);
            else 
                newfile();
            break;

        case 'M':
            dispmovblk();
            break;

        case 'N':
            namefile(" New file name:",file_name);
            break;

        case 'O':
            doscall();
            break;

        case 'P':
            inpfunckey();
            break;

        case 'Q':
            if (OK == saveyn()) retcode = ERR;
            break;

        case 'R':
            findrepl(1);
            break;

        case 'S':
            if (savefile()) beep(1);
            break;
 
        case 'T':
            settabs();
            break;

        case 'U':
            if (block_beg < block_end) {
                if ((block_beg >= scr_loff[row_no]) && 
                 (block_end <= scr_loff[row_no] + scr_llen[row_no])) {
                    setdraw();
                    draw_col = 0;
                }
                else
                    setdrawall();
            }
            block_beg = 0;
            block_end = 0;
            dispmess(" Block unmarked",-1,FALSE);
            break;

        case 'V':
            if (savecfg())
                dispmess(" Unable to save nledcfg.bin",-1,FALSE);
            else
                dispmess(" Configuration saved",-1,FALSE);
            break;

        case 'X':
            if (savefile()) beep(1);
            else retcode = ERR;
            break;

        case 'Y':
            setcase();
            break;

        case 'Z':
            dispchgcase();
            break;

        case '?':
            help();
            again = 1;
            break;

        case ESC:
        case LF:
        case CR:
            break;

        default:
            beep(1);
            again = 1;
            break;
        }
    }
    return(retcode);
}

/***********************************************************************\
* Undoes various NLED initializations. Returns OK if no problem, ERR    *
* if there was a fatal problem                                          *
* 11-Feb-1992: fixed screen size hardcoding                             *
\***********************************************************************/
int finalize(void)
{
    /* scroll entire screen up 1 (to get bright line off bottom) */
    qd_scrollup(0, num_rows - 1);

    /* move to the bottom row & close off fast screen display */
    qd_move(num_rows - 1,0);
    qd_flush();
    qd_close();

    /* deallocate space grabbed for data */
    datclose();

    return(OK);
}

/***********************************************************************\
* Displays help message and wait for a key to be pressed. Returns       *
* ESCHIT if escape was the key pressed, and OK otherwise.               *
* 01-APR-87: Removed unused var "c", and changed text                   *
* 28-APR-87: turn display off while writing screen                      *
* 22-APR-88: Added option Z                                             *
* 30-Dec-88: Changed copyright date                                     *
* 05-Oct-91: Changed copyright date                                     *
* 12-Feb-92: rearrangements, new case sensitivity option                *
* 04-Apr-94: Tab compression                                            *
* 09-Jul-95: Put line at bottom                                         *
* 02-Feb-02: Changed copyright date/message                             *
\***********************************************************************/
void help(void)
{
    int i;

    static char *help[] = {
"----------  NLED Programmer's Text Editor - Quick Online Reference  ----------",
"----------------- Copyright 1987-2002: Evan Alexander Weaver -----------------",
"<Esc> sequence summary: Press <Esc> to get the Command? prompt, then enter:",
"   A (*) to set Attention level (beep)  O to run an Operating system command",
"   B to mark Beginning of block         P (*) to Program a function key",
"   C to Copy block to current position  Q to Quit without saving file",
"   D to Delete block                    R to find and Replace text",
"   E to mark End of block               S to Save file to disk",
"   F to Find a text string              T (*) to set number of spaces per Tab",
"   G to Go to a specific line           U to Unmark block",
"   H (*) to cHange file access mode     V (*) to saVe configuration",
"   I to Import another file into this   X to save file and eXit",
"   K (*) to toggle tab c(K)ompression   Y (*) to set case sensitivitY for find",
"   L to Load a new file into NLED       Z to change case of current character",
"   M to Move block to current position  ? to see this screen",
"   N to give a new Name to current file <Esc> or <Enter> to go back",
"             (*) - affects configuration which currently is:",
"",
"------------------------------------------------------------------------------",
"Special keys (in addition to normal cursor movement):",
#if UNIX
" <Ctl/B>,<Ctl/E> Beg & End of line;  <Ctl/F>,<Ctl/L> First & Last screen line;",
" <Ctl/U>,<Ctl/D> Page Up & Down;  <Ctl/T>,<Ctl/R> Top & bottom (Rear) of file;",
" <Ctl/P>,<Ctl/N> Previous & Next word; <Ctl/X> Enter hex code; <Ctl/O> Ins/Ovr",
#else
" <Home>,<End> Beg & End of line;   <Ctl/Home>,<Ctl/End> Top & Bottom of screen;",
" <PgUp>,<PgDn> Previous & Next screen; <Ctl/PgUp>,<Ctl/PgDn> Beg & End of file;",
" <Ctl/Left>,<Ctl/Right> Previous & Next word;   <Ctl/Backspace> Enter hex code;",
#endif
"------------------------------------------------------------------------------"
};

    qd_off();
    /* only show as many rows as will fit on the screen */
    for (i = 0; i < num_rows - 1 && i < 24; i++){
        qd_move(i,0);
        printstr(help[i]);
        qd_eeol();
    }
    if (num_rows > 18) {
        qd_move(17,0);
        printstr(" Attention level:");
        printindex(beep_level,1);
        printstr("  Tab:");
        printindex(tab_size,2);
        printstr("[Compress O");
        printstr(tab_compress? "n] " : "ff]");
        printstr("  File mode:");
        printstr(text_mode?((text_mode==TEXT)?"Text  " : "ZText ") : "Binary");
        printstr("  Case Sensitive:");
        printstr(case_sens ? "Yes" : "No");
        qd_eeol();
    }

    qd_on();
    setdrawall();
}

/**********************************************************************\
* Performs various NLED initializations. Returns OK if no problem, ERR *
* if there was a fatal problem                                         *
* 01-APR-87: Also set "text_mode" and "ins_mode" if no config file     *
* 28-APR-87: turn display off while clearing screen                    *
* 27-DEC-88: removed 0 default for function keys                       *
* 17-Feb-92: check screen size, case sensitivity, colors, config_file  *
\**********************************************************************/
int initialize(void)
{
    /* initialization for fast display routines */
    qd_open();
    if ((num_rows = qd_rows()) > MAX_ROWS)
        num_rows = MAX_ROWS;
    num_cols = qd_cols();
    if (num_rows < 5 || num_cols < 40) {
        qd_close();
        puts("Screen is too small to run the Neat Little Editor\n");
        exit(1);
    }

    draw_brow = num_rows - 1;
    draw_col = num_cols;

    /* screen type setting may have been specified by the user */
    switch (scr_type) {
    case -1: /* use display settings determined by qd_open() */
        break;

    case 0: /* display that does not generate snow */
        qd_snowoff();
        break;

    case 1: /* display that does generate snow */
        qd_snowon();
        break;

    case 2: /* display that requires use of video ROM BIOS routines */
        qd_usebios();
        break;
    }

    /* get space for editing our file */
    if (datopen()) {
        printstr("There is not enough memory to run NLED\n");
        return(ERR);
    }

    /* don't autostart a function key macro */
    func_key = -1;

    /* read the configuration file, & set defaults if there's a problem */
    if (loadcfg(config_file)) {
        tab_size = 8;
        beep_level = 2;
        text_mode = ins_mode = TRUE;
    }
    if (color_setting > 0)
        qd_color(color_setting);

    /* clear the screen & display status line without a message */
    qd_off();
    qd_clear();
    qd_on();
    msg_flag = FALSE;
    dispstatus();

#ifdef SHAREWARE
    /* display shareware notice screen */
    shareware();
#endif

    /* load and display the file in file_name, if any */
    newfile();
    display();

    return(OK);
}

/*---------------------------END-OF-NLED.C----------------------------*/
